/***************************************************************************
Copyright (c) 2016, Guillaume Baud-Berthier

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
***************************************************************************/

extern "C" {
#include "aiger.h"
}

#include "aigprep.h"
#include "bmc.h"
#include "zigzag.h"

#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>

bool        cex       = 0;
bool        elim_ll   = 0;
bool        elim_uhl  = 0;
bool        elim_hl   = 0;
bool        bmc_only  = 0;
clock_t     clk_diff, clk;



SimpSolver* solver;
vec<Lit>    assumptions;

unsigned vars = 0;
std::map<unsigned, ands_def*> map_ands;
std::vector<idef*> mapdef;
std::vector<dll_list<ands_def>*> list_ands;

const char * name;
aiger * model, *model_coi;

int verbose_aigMC, bad_output;

static void SIG_handler(int signum) {
  printf("-1\nb0\n.\n");
  fflush(stdout);
  exit(0);
}

void init_aigMC() {
  model = aiger_init ();
}

void destroy_aigMC() {
  aiger_reset (model);
}

int isnum (const char * str) {
  const char * p = str;
  if (!isdigit (*p++)) return 0;
  while (isdigit (*p)) p++;
  return !*p;
}

const char * usage =
  "usage           : aigMC [option] <model> [<start> <stop>]\n"
  "-h              : help\n"
  "-v              : verbose\n"
  "-b              : use output as bad state\n"
  "-cex            : print counterexample\n"
  "-elim_ll        : perform Low Level VE\n"
  "-elim_uhl       : perform Unrolled High Level VE\n"
  "-elim_hl        : perform High Level VE\n"
  "-bmc            : perform BMC only\n"
  "start           : starting depth (Default: 0)\n"
  "stop            : max depth      (Default: INT_MAX)\n";

int main (int argc, char ** argv) {
  int i, j, k, start, stop;
  const char * err;
  start = -1;
  stop = -1;
  for (i = 1; i < argc; i++) {
    if (!strcmp (argv[i], "-h")) {
      printf ("%s", usage);
      exit (0);
    } else if (!strcmp (argv[i], "-v")) {
      verbose_aigMC++;
    } else if (!strcmp (argv[i], "-b")) {
      bad_output = 1;
    } else if (!strcmp (argv[i], "-cex")) {
      cex = 1;
    } else if (!strcmp (argv[i], "-bmc")) {
      bmc_only = 1;
    } else if (!strcmp (argv[i], "-elim_ll")) {
      elim_ll = 1;
    } else if (!strcmp (argv[i], "-elim_uhl")) {
      elim_uhl = 1;
    } else if (!strcmp (argv[i], "-elim_hl")) {
      elim_hl = 1;
    } else if (argv[i][0] == '-') {
      printf ("%s", usage);
      printf("Unknown option : '%s'\n", argv[i]);
      exit(-1);
    } else if (name && stop >= 0) {
      printf ("%s", usage);
      printf("Unknown option : '%s'\n", argv[i]);
      exit(-1);
    } else if (name && !isnum (argv[i])) {
      printf ("%s", usage);
      printf("Unknown option : '%s'\n", argv[i]);
      exit(-1);
    } else if (start < 0 && isnum (argv[i])) {
      start = atoi (argv[i]);
    } else if (stop < 0 && isnum (argv[i])) {
      stop = atoi (argv[i]);
    } else {
      name = argv[i];
    }
  }

  if (start < 0) start = 0;
  if (stop < 0) stop = INT_MAX;

  signal(SIGINT,  SIG_handler);
  signal(SIGTERM, SIG_handler);

  if (verbose_aigMC)
    printf("[MC]  AIGER MC from %d to %d\n", start, stop);
  init_aigMC();

  if (verbose_aigMC)
    printf("[MC]  Checking model : '%s'\n", name);

  
  if (name) err = aiger_open_and_read_from_file (model, name);
  if (err) {
    printf("Parsing error '%s': %s\n", name, err);
    destroy_aigMC();
    return -1;
  }

  if (model->num_justice) {
    printf("Justice properties are not supported yet!\n");
    exit(-1);
  }
  if (model->num_constraints) { 
    printf("Constraints are not supported yet!\n");
    exit(-1);
  }
  if (model->num_fairness) {
    printf("Fairness constraints are not supported yet!\n");
    exit(-1);
  }
  if (!model->num_bad) {
    if (!bad_output) {
      printf("No properties! (add -b to consider output as property)");
      exit(-1);
    } else {
      if (model->num_outputs > 1) {
	printf("Multiple properties are not supported yet!\n");
	exit(-1);

      } else {
	aiger_add_bad (model, model->outputs[0].lit, 0);
      }
    }
  }
  
  int saved_nb_latches = -1;
  if (cex)
    saved_nb_latches = model->num_latches;

  aiger_reencode (model);

  if (verbose_aigMC) printf("[aigprep]  Preprocessing\n");
  destroy_prep();
  init_prep(model, cex, verbose_aigMC);
  iter_defs(cex);
  model = rebuild_aig(true);
  destroy_prep();

  if (verbose_aigMC) {
    printf("[MC]  MILOA = %u %u %u %u %u\n",
	   model->maxvar,
	   model->num_inputs,
	   model->num_latches,
	   model->num_outputs,
	   model->num_ands);
    
    printf("[MC]  BCJF = %u %u %u %u\n",
	   model->num_bad,
	   model->num_constraints,
	   model->num_justice,
	   model->num_fairness);
  }
  
  if(model->bad[0].lit < 2) {
    if (!model->bad[0].lit) {
      if (verbose_aigMC) printf("[MC]  Bad state is equal to FALSE\n");
      printf("0\nb0\n.\n");
    } else {
      if (verbose_aigMC) printf("[MC]  Bad state is equal to TRUE\n");
      printf("1\nb0\n.\n");
    }

  } else {
    try {
      if (bmc_only) {
	if (verbose_aigMC) printf("[MC]  Starting BMC\n");
	bmc(start, stop, model, elim_ll, elim_uhl, elim_hl, cex, saved_nb_latches, verbose_aigMC);
      } else {
	zigzag(start, stop, model, elim_ll, elim_uhl, elim_hl, /* simple */ 1, verbose_aigMC);
      }
    } 
    catch(OutOfMemoryException&) {
      printf("-1\nb0\n.\n");
    }
    catch(std::bad_alloc&) {
      printf("-1\nb0\n.\n");
    }
  }


  return 0;
}
