/***************************************************************************
Copyright (c) 2016, Guillaume Baud-Berthier

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
***************************************************************************/

#include "bmc.h"

using namespace Glucose;

unsigned UNDEF_LIT = INT_MAX;

float enco_cumul_time = 0;


/* ========================================================================== */
/*                           BMC algorithm functions                          */
/* ========================================================================== */

int bmc(int start, int stop, aiger *model, bool elim_ll, bool elim_uhl,
	bool elim_hl, bool cex, int nb_latches, bool verbose_) {
  int i, j;
  int status = -1;
  clock_t clk;


  if (verbose_) clk = clock();

  if (verbose_) printf("Mem: %d MB\n", get_mem());

  unsigned bad_lit = init_model(model, stop, elim_ll, verbose_);

  if (elim_hl)
    hl_ve(verbose_);

  if (verbose_) {
    clk = clock() - clk;
    enco_cumul_time += (float)(clk)/(float)(CLOCKS_PER_SEC);
  }

  for(i = 0; i <= stop; i++) {

    if (verbose_) printf("Mem: %d MB\n", get_mem());
    if (verbose_) printf("[bmc]  Encoding %d..\n", i);
    if (verbose_) clk = clock();

    list_ands.push_back(unroll(i));
    
    iter_flatten_ptu(i);

    if (elim_uhl) {
      // Disable PO varelim before performing unrolled high level varelim
      unsigned var_po = mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i];
      if (var_po > 1) {
	std::map<unsigned, ands_def*>::iterator it_lit_ref = map_ands.find(aiger_lit2var(var_po));
  
	assert(it_lit_ref != map_ands.end()); // Might be false if po <=> input
	assert(it_lit_ref->second->cst == -1);

	it_lit_ref->second->apply_ve = 0;
      }

      
      un_hl_ve(verbose_, i);
    }

    if (elim_ll) {
      if (verbose_) printf("[bmc]  Warning : k-COI is disabled because of varelim!\n");
      ands2clauses(i);
    } else {
      ands2clauses_coi(bad_lit, i, verbose_);
    }

    if (verbose_) {
      clk = clock() - clk;
      enco_cumul_time += (float)(clk)/(float)(CLOCKS_PER_SEC);
    }
    
    if (verbose_) printf("[bmc]  Solving..\n");

    if ((!mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i] &&
	 !aiger_sign(bad_lit)) ||
	(mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i] == 1 &&
	 aiger_sign(bad_lit))) {
      
      if (verbose_) printf("[bmc]  UNSAT (Solved by simplification)\n");
      printf("u%d\n", i);
      fflush(stdout);

      if (!dll_size(dll_latches)) {
	status = 0;
	break;
      }

    } else if ((!mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i] &&
		aiger_sign(bad_lit)) ||
	       (mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i] == 1 &&
		!aiger_sign(bad_lit))) {

      if (verbose_) printf("[bmc]  CEX (Solved by simplification)\n");
      status = 1;
      break;
	
    } else {

      if (verbose_) clk = clock();

      unsigned var_po = mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i];
      assert(var_po > 1);
      std::map<unsigned, ands_def*>::iterator it_lit_ref = map_ands.find(aiger_lit2var(var_po));

      assert(it_lit_ref != map_ands.end()); // Might be false if po <=> input
      assert(it_lit_ref->second->cst == -1);
      
      int lit_bad = addPO(bad_lit, i);

      assume(mapdef[0]->lhs_ts.back() >> 1);



      /* Freeze latch variables */
      if (elim_ll) {
	dll_elem<idef> *tmp_e = dll_latches->head;
	idef* tmp_def;
	while(tmp_e != NULL) {
	  tmp_def = tmp_e->def;

	  assert(tmp_def->rhs.size() == 1);
	  if (tmp_def->rhs.front() > 1) {
	    int next_lit = (mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts.back());

	    /* Temporary fix.. */
	    it_lit_ref = map_ands.find(aiger_lit2var(next_lit));
	    if (it_lit_ref != map_ands.end()) {
	      if (it_lit_ref->second->cst != -1) {
		next_lit = it_lit_ref->second->cst;
		if (it_lit_ref->second->cst_ref != NULL)
		  assert(it_lit_ref->second->cst_ref->cst == -1); // Could be false.. [TODO]
	      }
	    }
	    /*******/

	    if (next_lit > 1) {
	      solver->frozen[next_lit >> 1] = true;
	    }
	  }

	  tmp_e = tmp_e->next;
	}
      }
      /**************************/
      if (verbose_) {
	clk = clock() - clk;
	enco_cumul_time += (float)(clk)/(float)(CLOCKS_PER_SEC);
      }
      
      lbool res_check;
      res_check = check_sat(verbose_);
      if (res_check == l_True) {
	status = 1;
	break;
      } else if (res_check == l_Undef) {
	assert(0);
      }


      printf("u%d\n", i);
      fflush(stdout);

      if (!dll_size(dll_latches)) {
	status = 0;
	break;
      }      
      
      /* Unfreeze latch variables */
      if (elim_ll)
	for(int l = 0; l < solver->frozen.size(); l++)
	  solver->frozen[l] = false;
      
      /**************************/      
      

      ucl(-(mapdef[0]->lhs_ts.back() >> 1)); //      
    }
  }

  
  printf("%d\nb0\n", status);
  fflush(stdout);
  if (cex && status == 1) {
    idef* tmp_def;

    for (int i = 0; i < nb_latches; i++)
      printf("0");
    printf("\n");

    for(j = 0; j <= i; j++) {
      dll_elem<idef> *tmp_e = dll_inputs->head;
      while(tmp_e != NULL) {
	tmp_def = tmp_e->def;

	assert(tmp_def->lhs_ts[j] > 1);

	if (solver->modelValue(tmp_def->lhs_ts[j] >> 1) == l_True)
	  printf("1");
	else if (solver->modelValue(tmp_def->lhs_ts[j] >> 1) == l_False)
	  printf("0");
	else
	  printf("x");

	tmp_e = tmp_e->next;
      }
      printf("\n");
    }
  }
  printf(".\n");

  if (verbose_) {
    printf("Encoding %.2fs | Simplification %.2fs | Solving %.2fs\n", enco_cumul_time,
	   solver->simp_cumul_time, solver->solv_cumul_time);
    if (elim_ll)
      printf("Vars %d (%d - %d - %d)| Clauses %d\n",
	     solver->nVars() - solver->eliminated_vars, solver->nVars(), get_stats_un_hl_ve(), solver->eliminated_vars, solver->nClauses());
    else
      printf("Vars %d (%d - %d - %d)| Clauses %d\n",
	     solver->nVars() - get_stats_un_hl_ve() - solver->eliminated_vars, solver->nVars(), get_stats_un_hl_ve(), solver->eliminated_vars, solver->nClauses());
  }
  fflush(stdout);

  if (verbose_) printf("Mem: %d MB\n", get_mem());

  destroy_model(verbose_);
  return 0;
}


