/***************************************************************************
Copyright (c) 2016, Guillaume Baud-Berthier

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
***************************************************************************/

#include "zigzag.h"

unsigned act_init;

// Simple path (loop free) constraints on demand
bool simple_path_od(int time, int bad, bool verbose_) {

  vec<Lit> lits;
  idef* tmp_def;
  dll_elem<idef> *tmp_e;
  bool same_state;
  bool run_solver_again = 1;

  while (run_solver_again) {
    run_solver_again = 0;

    for (int i = 0; i < time && !run_solver_again; i++) {
      for (int j = i+1; j <= time && !run_solver_again; j++) {
	
	same_state = 1;

	tmp_e = dll_latches->head;
	while(tmp_e != NULL && same_state) {
	  tmp_def = tmp_e->def;

	  //assert(tmp_def->lhs_ts[i] > 1);
	  //assert(tmp_def->lhs_ts[j] > 1);
	  assert(solver->model[tmp_def->lhs_ts[i] >> 1] != l_Undef);
	  assert(solver->model[tmp_def->lhs_ts[j] >> 1] != l_Undef);

	  if (tmp_def->lhs_ts[i] > 1 && tmp_def->lhs_ts[j] > 1) {
	    assert(!solver->isEliminated(tmp_def->lhs_ts[i] >> 1));
	    assert(!solver->isEliminated(tmp_def->lhs_ts[j] >> 1));
	    
	    if (((solver->model[tmp_def->lhs_ts[i] >> 1] != solver->model[tmp_def->lhs_ts[j] >> 1]) &&
		 (aiger_sign(tmp_def->lhs_ts[i]) == aiger_sign(tmp_def->lhs_ts[j]))) || 
		((solver->model[tmp_def->lhs_ts[i] >> 1] == solver->model[tmp_def->lhs_ts[j] >> 1]) &&
		 (aiger_sign(tmp_def->lhs_ts[i]) != aiger_sign(tmp_def->lhs_ts[j]))))
	      same_state = 0;

	  } else if (tmp_def->lhs_ts[i] > 1) {
	    assert(!solver->isEliminated(tmp_def->lhs_ts[i] >> 1));
	    if ((tmp_def->lhs_ts[j] == 1 && !aiger_sign(tmp_def->lhs_ts[i])) ||
		(tmp_def->lhs_ts[j] == 0 &&  aiger_sign(tmp_def->lhs_ts[i]))) {
	      same_state = solver->model[tmp_def->lhs_ts[i] >> 1] == l_True;
	    } else {
	      same_state = solver->model[tmp_def->lhs_ts[i] >> 1] == l_False;
	    }

	  } else if (tmp_def->lhs_ts[j] > 1) {
	    assert(!solver->isEliminated(tmp_def->lhs_ts[j] >> 1));
	    if ((tmp_def->lhs_ts[i] == 1 && !aiger_sign(tmp_def->lhs_ts[j])) ||
		(tmp_def->lhs_ts[i] == 0 &&  aiger_sign(tmp_def->lhs_ts[j]))) {
	      same_state = solver->model[tmp_def->lhs_ts[j] >> 1] == l_True;
	    } else {
	      same_state = solver->model[tmp_def->lhs_ts[j] >> 1] == l_False;
	    }

	  } else {
	    same_state = tmp_def->lhs_ts[i] == tmp_def->lhs_ts[j];
	  }

	  tmp_e = tmp_e->next;
	}

	if (same_state) {
	  // Add constraints
	  lits.clear();
	  tmp_e = dll_latches->head;
	  while(tmp_e != NULL && same_state) {
	    tmp_def = tmp_e->def;

	    if (tmp_def->lhs_ts[i] > 1 && tmp_def->lhs_ts[j] > 1) {
	      //assert(tmp_def->lhs_ts[i] > 1);
	      //assert(tmp_def->lhs_ts[j] > 1);


	      unsigned new_v = new_var() >> 1;
	      while (new_v >= solver->nVars()) solver->newVar();

	      //assert(!aiger_sign(tmp_def->lhs_ts[i]) && !aiger_sign(tmp_def->lhs_ts[j]));
	      int a = aiger_sign(tmp_def->lhs_ts[i]) ? -(tmp_def->lhs_ts[i] >> 1) : tmp_def->lhs_ts[i] >> 1;
	      int b = aiger_sign(tmp_def->lhs_ts[j]) ? -(tmp_def->lhs_ts[j] >> 1) : tmp_def->lhs_ts[j] >> 1;

	      lits.push(mkLit(new_v));

	      // tcl(-new_v, -(tmp_def->lhs_ts[i] >> 1), -(tmp_def->lhs_ts[j] >> 1));
	      // tcl(-new_v,  (tmp_def->lhs_ts[i] >> 1),  (tmp_def->lhs_ts[j] >> 1));
	      // tcl( new_v,  (tmp_def->lhs_ts[i] >> 1), -(tmp_def->lhs_ts[j] >> 1));
	      // tcl( new_v, -(tmp_def->lhs_ts[i] >> 1),  (tmp_def->lhs_ts[j] >> 1));

	      tcl(-new_v, -a, -b);
	      tcl(-new_v,  a,  b);
	      tcl( new_v,  a, -b);
	      tcl( new_v, -a,  b);

	    } else {
	      // TODO:
	      assert(0);
	    }

	    tmp_e = tmp_e->next;
	  }
	
	  solver->addClause_(lits);
	  run_solver_again = 1;
	}
      }
    }

    if (run_solver_again) {
      assume(bad); // Assume: -P(x)
      if (check_sat(verbose_) == l_False)
	return 1;
    }
  }

  return 0;
}


// // Simple path (loop free) constraints on demand
// bool simple_path_od(int time, int bad, bool verbose_) {

//   vec<Lit> lits;
//   idef* tmp_def;
//   dll_elem<idef> *tmp_e;
//   bool same_state;
//   bool run_solver_again = 1;

//   while (run_solver_again) {
//     run_solver_again = 0;

//     for (int i = 0; i < time && !run_solver_again; i++) {
//       for (int j = i+1; j <= time && !run_solver_again; j++) {
	
// 	same_state = 1;

// 	tmp_e = dll_latches->head;
// 	while(tmp_e != NULL && same_state) {
// 	  tmp_def = tmp_e->def;

// 	  //assert(tmp_def->lhs_ts[i] > 1);
// 	  //assert(tmp_def->lhs_ts[j] > 1);
// 	  if (tmp_def->lhs_ts[i] > 1 && tmp_def->lhs_ts[j] > 1) {
// 	    assert(!solver->isEliminated(tmp_def->lhs_ts[i] >> 1));
// 	    assert(!solver->isEliminated(tmp_def->lhs_ts[j] >> 1));
	    
// 	    if (solver->model[tmp_def->lhs_ts[i] >> 1] != solver->model[tmp_def->lhs_ts[j] >> 1])
// 	      same_state = 0;
	    
// 	  } else if (tmp_def->lhs_ts[i] > 1) {
// 	    if (tmp_def->lhs_ts[j] == 1) {
// 	      same_state = solver->model[tmp_def->lhs_ts[i] >> 1] == l_True;
// 	    } else {
// 	      same_state = solver->model[tmp_def->lhs_ts[i] >> 1] == l_False;
// 	    }
// 	  } else if (tmp_def->lhs_ts[j] > 1) {
// 	    if (tmp_def->lhs_ts[i] == 1) {
// 	      same_state = solver->model[tmp_def->lhs_ts[j] >> 1] == l_True;
// 	    } else {
// 	      same_state = solver->model[tmp_def->lhs_ts[j] >> 1] == l_False;
// 	    }
// 	  } else {
// 	    same_state = tmp_def->lhs_ts[i] == tmp_def->lhs_ts[j];
// 	  }

// 	  tmp_e = tmp_e->next;
// 	}

// 	if (same_state) {
// 	  // Add constraints
// 	  lits.clear();
// 	  tmp_e = dll_latches->head;
// 	  while(tmp_e != NULL && same_state) {
// 	    tmp_def = tmp_e->def;


// 	    //assert(tmp_def->lhs_ts[i] > 1);
// 	    //assert(tmp_def->lhs_ts[j] > 1);
// 	    if (tmp_def->lhs_ts[i] > 1 && tmp_def->lhs_ts[j] > 1) {

// 	    unsigned new_v = new_var() >> 1;
// 	    while (new_v >= solver->nVars()) solver->newVar();

// 	    //assert(!aiger_sign(tmp_def->lhs_ts[i]) && !aiger_sign(tmp_def->lhs_ts[j]));
// 	    int a = aiger_sign(tmp_def->lhs_ts[i]) ? -(tmp_def->lhs_ts[i] >> 1) : tmp_def->lhs_ts[i] >> 1;
// 	    int b = aiger_sign(tmp_def->lhs_ts[j]) ? -(tmp_def->lhs_ts[j] >> 1) : tmp_def->lhs_ts[j] >> 1;

// 	    lits.push(mkLit(new_v));

// 	    tcl(-new_v, -a, -b);
// 	    tcl(-new_v,  a,  b);
// 	    tcl( new_v,  a, -b);
// 	    tcl( new_v, -a,  b);

// 	    } else {
// 	      // TODO:
// 	      assert(0);
// 	    }

// 	    tmp_e = tmp_e->next;
// 	  }
	
// 	  solver->addClause_(lits);
// 	  run_solver_again = 1;
// 	}
//       }
//     }

//     if (run_solver_again) {
//       assume(bad); // Assume: -P(x)
//       if (check_sat(verbose_) == l_False)
// 	return 1;
//     }
//   }

//   return 0;
// }

dll_list<ands_def>* unroll_kind(int time) {
  int j;
  unsigned tmp_rhs;
  unsigned tmp_rhs0, tmp_rhs1;

  dll_list<ands_def> *tmp_list_ands;
  tmp_list_ands = new dll_list<ands_def>();
  dll_init<ands_def>(tmp_list_ands);
  
  idef* tmp_def;
  dll_elem<idef> *tmp_e;
  std::list<unsigned>::iterator it_rhs;
  std::set<unsigned>::iterator it_tmp_rhs;


  // Inputs
  tmp_e = dll_inputs->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;
    
    tmp_def->lhs_ts.push_back(new_var());
    tmp_e = tmp_e->next;
  }  
  
  // Latches
  tmp_e = dll_latches->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;

    if (!time) { // Init
      tmp_def->lhs_ts.push_back(new_var());

      bcl(-(tmp_def->lhs_ts[0] >> 1), -act_init);

    } else {
      assert(tmp_def->rhs.size() == 1);
      
      if (tmp_def->rhs.front() > 1) {
	assert(tmp_def->lhs_ts[time-1] != -1);

	// assert(tmp_def->lhs_ts.size() == time);
	// tmp_def->lhs_ts.push_back(new_var());

	// if (aiger_sign(tmp_def->rhs.front()) != aiger_sign(mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1])) {
	//   bcl( (tmp_def->lhs_ts[time] >> 1),  (mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1));
	//   bcl(-(tmp_def->lhs_ts[time] >> 1), -(mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1));
	// } else {
	//   bcl( (tmp_def->lhs_ts[time] >> 1), -(mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1));
	//   bcl(-(tmp_def->lhs_ts[time] >> 1),   mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1);
	// }
	if (aiger_sign(tmp_def->rhs.front()))
	  tmp_def->lhs_ts.push_back
	    (aiger_not(mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1]));
	else
	  tmp_def->lhs_ts.push_back
	    (mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1]);

      } else { // Cst latch
	tmp_def->lhs_ts.push_back(new_var());
	assert(tmp_def->rhs.front() == 1);
	ucl(tmp_def->lhs_ts[time] >> 1);
      }
    }
    tmp_e = tmp_e->next;
  }
  
  // Ands
  tmp_e = dll_ands->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;

    prop_tr_un *tpu = unroll_prop_tr(tmp_def->rhs_tr, time, 0);

    //assert(tpu->lit == UNDEF_LIT);
    if (tpu->lit == UNDEF_LIT) {
      assert(tpu->op != VAR);
    
      tmp_def->lhs_ts.push_back(new_var());
    
      ands_def *tmp_ands = new ands_def(tmp_def->lhs_ts[time], tpu, tmp_def, time);
    
      map_ands.insert(std::pair<unsigned, ands_def*>(aiger_lit2var(tmp_ands->lhs), tmp_ands));
    
      tmp_ands->dllade = dll_insert_last(tmp_list_ands, aiger_lit2var(tmp_def->lhs_ts[time]), tmp_ands);

    } else {
      assert(tpu->op == VAR);
      tmp_def->lhs_ts.push_back(tpu->lit);
    }

    tmp_e = tmp_e->next;
  }

  return tmp_list_ands;
}





int zigzag(int start, int stop, aiger *model, bool elim_ll, bool elim_uhl,
	   bool elim_hl, int simple, bool verbose_) {
  int i, j;

  int status = -1;

  unsigned bad_lit = init_model(model, stop, elim_ll, verbose_);

  if (elim_hl)
    hl_ve(verbose_);

  act_init = new_var() >> 1;

  for(i = 0; i <= stop; i++) {

    if (verbose_) printf("[zigzag]  Encoding %d..\n", i);

    list_ands.push_back(unroll_kind(i));

    iter_flatten_ptu(i);

    if (elim_uhl) {
      // Disable PO varelim before performing unrolled high level varelim
      unsigned var_po = mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i];
      if (var_po > 1) {
	std::map<unsigned, ands_def*>::iterator it_lit_ref = map_ands.find(aiger_lit2var(var_po));

	//assert(it_lit_ref != map_ands.end()); // Might be false if po <=> input
	if (it_lit_ref != map_ands.end()) {
	  assert(it_lit_ref->second->cst == -1);
	  
	  it_lit_ref->second->apply_ve = 0;
	}
      }

      un_hl_ve_zz(verbose_, i);
    }

    ands2clauses(i);

    assert(mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i] > 1);
    
    int bad = mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i];
    bad = aiger_sign(bad_lit) == aiger_sign(bad) ? (bad >> 1) : -(bad >> 1);
    

    if (elim_ll) {
      /* Freeze latch variables : init */
      if(!i) {
	dll_elem<idef> *tmp_e = dll_latches->head;
	idef* tmp_def;
	while(tmp_e != NULL) {
	  tmp_def = tmp_e->def;
	  
	  assert(tmp_def->rhs.size() == 1);
	  int next_lit = tmp_def->lhs_ts[0];
      
	  solver->frozen[next_lit >> 1] = true;

	  tmp_e = tmp_e->next;
	}
      }

      /* Freeze latch variables : next */
      dll_elem<idef> *tmp_e = dll_latches->head;
      idef* tmp_def;
      while(tmp_e != NULL) {
	tmp_def = tmp_e->def;

	assert(tmp_def->rhs.size() == 1);
	if (tmp_def->rhs.front() > 1) {
	  int next_lit = (mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts.back());

	  /* Temporary fix.. */
	  std::map<unsigned, ands_def*>::iterator it_lit_ref = map_ands.find(aiger_lit2var(next_lit));
	  if (it_lit_ref != map_ands.end()) {
	    if (it_lit_ref->second->cst != -1) {
	      next_lit = it_lit_ref->second->cst;
	      if (it_lit_ref->second->cst_ref != NULL)
		assert(it_lit_ref->second->cst_ref->cst == -1); // Could be false.. [TODO]
	    }
	  }
	  /*******/
	  
	  if (next_lit > 1) {
	    solver->frozen[next_lit >> 1] = true;
	  }
	}

	tmp_e = tmp_e->next;
      }

      /* Freeze assumption variables */
      solver->frozen[act_init] = true;
      solver->frozen[abs(bad)] = true; // Not required, assumptions are frozen
    }


    lbool res_check;

    if (verbose_) printf("[zigzag]  Solving %d-induction..\n", i);
      
    assume(bad); // Assume: -P(x)
    
    res_check = check_sat(verbose_);
    if (res_check == l_False) {
      status = 0;
      break;
    } else if (i > 1 && simple) {
      if (simple_path_od(i, bad, verbose_)) {
	status = 0;
	break;
      }
    }


    if (verbose_) printf("[zigzag]  Solving BMC %d..\n", i);

    assume(bad); // Assume: -P(x)
    assume(act_init);                  // Assume:  I(x)

    res_check = check_sat(verbose_);
    if (res_check == l_True) {
      status = 1;
      break;
    }


    printf("u%d\n", i);
    fflush(stdout);

    ucl(-bad);                   // Add: P(x)

  }

  printf("%d\nb0\n.\n", status);
  fflush(stdout);

  if (verbose_) {
      printf("Vars %d (%d - %d)| Clauses %d\n", solver->nVars() - solver->eliminated_vars,
	     solver->nVars(), solver->eliminated_vars, solver->nClauses());
  }

  destroy_model(verbose_);
  return 0;
}

